/* sane - Scanner Access Now Easy.
   Copyright (C) 1996, 1997 David Mosberger-Tang
   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file provides a generic SCSI interface.  */

#include <sane/config.h>

#include <lalloca.h>	/* MUST come first for AIX! */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/param.h>

#define STUBBED_INTERFACE	0
#define LINUX_INTERFACE		1
#define BSD_INTERFACE		2
#define	HPUX_INTERFACE		3
#define OPENSTEP_INTERFACE	4
#define DECUNIX_INTERFACE	5

#ifdef HAVE_SCSI_SG_H
# define USE LINUX_INTERFACE
# include <scsi/sg.h>
#else
# ifdef HAVE__USR_SRC_LINUX_INCLUDE_SCSI_SG_H
#  define USE LINUX_INTERFACE
#  include "/usr/src/linux/include/scsi/sg.h"
# else
#  ifdef HAVE_SYS_SCSI_H
#   include <sys/scsi.h>
#   ifdef SCTL_READ
#    define USE HPUX_INTERFACE
#   else
     /* This happens for AIX... */
#   endif
#  else
#   ifdef HAVE_SYS_SCSIIO_H
#    define USE BSD_INTERFACE
#    include <sys/scsiio.h>
#    ifdef HAVE_SCSI_H
#     include <scsi.h>
#    endif
#   else
#    ifdef HAVE_BSD_DEV_SCSIREG_H
#     define USE OPENSTEP_INTERFACE
#     include <bsd/dev/scsireg.h>
#    else
#     ifdef HAVE_IO_CAM_CAM_H
#      define USE DECUNIX_INTERFACE
#      include <io/common/iotypes.h>
#      include <io/cam/cam.h>
#      include <io/cam/dec_cam.h>
#      include <io/cam/uagt.h>
#      include <io/cam/scsi_all.h>
#     endif
#    endif
#   endif
#  endif
# endif
#endif

#ifndef USE
# define USE STUBBED_INTERFACE
#endif

#include <sane/sanei_scsi.h>
#include <sane/sanei.h>

#define BACKEND_NAME	sanei_scsi
#include <sane/sanei_debug.h>


#ifdef SG_BIG_BUFF
int sanei_scsi_max_request_size = SG_BIG_BUFF;
#else
# ifdef __FreeBSD__
int sanei_scsi_max_request_size =  32 * 1024; /* FreeBSD is limited to this */
# else
int sanei_scsi_max_request_size = 120 * 1024; /* works at least for OpenStep */
# endif
#endif

static int cam_fd = -1;		/* used for SCSI CAM based interfaces */
static struct
  {
    u_int in_use  : 1;		/* is this fd_info in use? */
    u_int fake_fd : 1;		/* is this a fake file descriptor? */
    u_int bus, target, lun;	/* nexus info; used for some interfaces only */
    SANEI_SCSI_Sense_Handler sense_handler;
  }
*fd_info;

static u_char cdb_sizes[8] = {6, 10, 10, 12, 12, 12, 10, 10};
#define CDB_SIZE(opcode)	cdb_sizes[(((opcode) >> 5) & 7)]

SANE_Status
sanei_scsi_open (const char * dev, int * fdp, SANEI_SCSI_Sense_Handler handler)
{
  static int num_alloced = 0;
  u_int bus = 0, target = 0, lun = 0, fake_fd = 0;
  int fd;

  DBG_INIT();

#if USE == DECUNIX_INTERFACE
  {
    UAGT_CAM_SCAN cam_scan;

    if (sscanf (dev, "b%dt%dl%d", &bus, &target, &lun) != 3)
      {
	DBG(1, "sanei_scsi_open: device name %s is not a valid\n",
	    strerror (errno));
	return SANE_STATUS_INVAL;
      }

    if (cam_fd < 0)
      {
	cam_fd = open ("/dev/cam", O_RDWR);
	if (cam_fd < 0)
	  {
	    DBG(1, "sanei_scsi_open: open(/dev/cam) failed: %s\n",
		strerror (errno));
	    return SANE_STATUS_INVAL;
	  }
      }
    cam_scan.ucs_bus = bus;
    cam_scan.ucs_target = target;
    cam_scan.ucs_lun = lun;
    if (ioctl (cam_fd, UAGT_CAM_SINGLE_SCAN, &cam_scan) < 0)
      {
	DBG(1, "sanei_scsi_open: ioctl(UAGT_CAM_SINGLE_SCAN) failed: %s\n",
	    strerror (errno));
	return SANE_STATUS_INVAL;
      }

    for (fd = 0; fd < num_alloced; ++fd)
      if (!fd_info[fd].in_use)
	break;
  }
#else /* !DECUNIX_INTERFACE */
# ifdef SGIOCSTL
  {
    char * real_dev;
    size_t len;

    /* OpenStep is a bit broken in that the device name refers to a scsi
       _bus_, not an individual scsi device.  Hence, SANE has to fudge
       with the device name so we know which target to connect to.  For
       this purpose, we use the last character in the device name as the
       target index.  'a' is target 0, 'b', target 1, and so on... */

    len = strlen (dev);
    if (len <= 1)
      {
	DBG(1, "sanei_scsi_open: devicename `%s' too short\n", dev);
	return SANE_STATUS_INVAL;
      }

    real_dev = strdup (dev);
    real_dev[len - 1] = '\0';

    target = dev[len - 1] - 'a';
    if (target > 7)
      {
	DBG(1, "sanei_scsi_open: `%c' is not a valid target id\n",
	    dev[len - 1]);
	return SANE_STATUS_INVAL;
      }
    dev = real_dev;
  }
# endif /* SGIOCSTL */

  fd = open (dev, O_RDWR | O_EXCL);
  if (fd < 0)
    {
      DBG(1, "sanei_scsi_open: open() failed: %s\n", strerror (errno));
      return SANE_STATUS_INVAL;
    }

# ifdef SG_SET_TIMEOUT
  /* Set large timeout since some scanners are slow but do not
     disconnect... ;-( */
  {
    int timeout;
    timeout = 10*60*HZ;		/* how about 10 minutes? ;-) */
    ioctl(fd, SG_SET_TIMEOUT, &timeout);
  }
# endif

# ifdef SGIOCSTL
  {
    struct scsi_adr sa;

    free ((void *) dev);

    sa.sa_target = target;
    sa.sa_lun = 0;
    if (ioctl (fd, SGIOCSTL, &sa) == -1)
      {
	DBG(1, "sanei_scsi_open: failed to attach to target: %u (%s)\n",
	    sa.sa_target, strerror (errno));
	return SANE_STATUS_INVAL;
      }
  }
# endif /* SGIOCSTL */
#endif /* !DECUNIX_INTERFACE */

  if (fd >= num_alloced)
    {
      size_t new_size, old_size;

      old_size = num_alloced * sizeof (fd_info[0]);
      num_alloced = fd + 8;
      new_size = num_alloced * sizeof (fd_info[0]);
      if (fd_info)
	fd_info = realloc (fd_info, new_size);
      else
	fd_info = malloc (new_size);
      memset ((char *) fd_info + old_size, 0, new_size - old_size);
      if (!fd_info)
	{
	  if (!fake_fd)
	    close (fd);
	  return SANE_STATUS_NO_MEM;
	}
    }
  fd_info[fd].sense_handler = handler;
  fd_info[fd].fake_fd = fake_fd;
  fd_info[fd].bus = bus;
  fd_info[fd].target = target;
  fd_info[fd].lun = lun;

  if (fdp)
    *fdp = fd;
  return SANE_STATUS_GOOD;
}

void
sanei_scsi_close (int fd)
{
  fd_info[fd].in_use = 0;
  fd_info[fd].sense_handler = 0;
  if (!fd_info[fd].fake_fd)
    close (fd);
}


#if USE == LINUX_INTERFACE
SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  /* The Linux generic scsi interface is braindead as it forces us to
     copy things around like mad.  Fortunately, most command requests
     are small.  */
  struct
    {
      struct sg_header	hdr;
      u_int8_t		data[SG_BIG_BUFF];
    }
  *req;
  ssize_t nread;

  assert (dst_size && *dst_size
	  ? CDB_SIZE (*(u_char *) src) == src_size
	  : CDB_SIZE (*(u_char *) src) <= src_size);

  req = alloca (sizeof (*req));	/* works around gdb bug... ;-( */

  memset (&req->hdr, 0, sizeof (req->hdr));
  req->hdr.pack_len  = src_size + sizeof (req->hdr);
  req->hdr.reply_len = (dst_size ? *dst_size : 0) + sizeof (req->hdr);
  memcpy (&req->data, src, src_size);

  if (write (fd, req, req->hdr.pack_len) != req->hdr.pack_len)
    {
      DBG(1, "sanei_scsi_cmd: bad write (errno=%s)\n", strerror (errno));
      return SANE_STATUS_IO_ERROR;
    }

  nread = read (fd, req, req->hdr.reply_len);
  if (nread < 0)
    {
      DBG(1, "sanei_scsi_cmd: read returned %ld (errno=%d)\n",
	  (long) nread, errno);
      return SANE_STATUS_IO_ERROR;
    }

  nread -= sizeof (req->hdr);

  if (req->hdr.result != 0)
    {
      DBG(1, "sanei_scsi_cmd: SCSI command failed: %s\n",
	  strerror (req->hdr.result));
      if (req->hdr.result == EBUSY)
	return SANE_STATUS_DEVICE_BUSY;
      else if ((req->hdr.sense_buffer[0] & 0x80) && fd_info[fd].sense_handler)
	return (*fd_info[fd].sense_handler) (fd, req->hdr.sense_buffer);
      else
	return SANE_STATUS_IO_ERROR;
    }

  if (dst)
    memcpy(dst, req->data, nread);

  if (dst_size)
    *dst_size = nread;

  return SANE_STATUS_GOOD;
}
#endif /* USE == LINUX_INTERFACE */


#if USE == BSD_INTERFACE

# ifndef HAVE_SCSIREQ_ENTER
static int
scsireq_enter (int fd, scsireq_t *hdr)
{
  return ioctl (fd, SCIOCCOMMAND, hdr);
}
# endif /* !HAVE_SCSIREQ_ENTER */

SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  size_t cdb_size;
  scsireq_t hdr;
  int result;

  cdb_size = CDB_SIZE (*(u_char *) src);

  memset (&hdr, 0, sizeof (hdr));
  memcpy (hdr.cmd, src, cdb_size);
  if (dst_size && *dst_size)
    {
      assert (cdb_size == src_size);
      hdr.flags = SCCMD_READ;
      hdr.databuf = dst;
      hdr.datalen = *dst_size;
    }
  else
    {
      assert (cdb_size <= src_size);
      hdr.flags = SCCMD_WRITE;
      hdr.databuf = (char *) src + cdb_size;
      hdr.datalen = src_size;
    }
  hdr.timeout = 60000;			/* 1 minute timeout */
  hdr.cmdlen = cdb_size;
  hdr.senselen = sizeof (hdr.sense);

  result = scsireq_enter (fd, &hdr);
  if (result < 0)
    {
      DBG(1, "sanei_scsi_cmd: scsi_reqenter() failed: %s\n", strerror (errno));
      return SANE_STATUS_IO_ERROR;
    }
  if (hdr.retsts != SCCMD_OK)
    {
      DBG(1, "sanei_scsi_cmd: scsi returned with status %d\n", hdr.retsts);
      switch (hdr.retsts)
	{
	case SCCMD_TIMEOUT:
	case SCCMD_BUSY:
	  return SANE_STATUS_DEVICE_BUSY;

	case SCCMD_SENSE:
	  if (fd_info[fd].sense_handler)
	    return (*fd_info[fd].sense_handler) (fd, &hdr.sense[0]);
	  /* fall through */
	default:
	  return SANE_STATUS_IO_ERROR;
	}
    }

  if (dst_size)
    *dst_size = hdr.datalen_used;

  return SANE_STATUS_GOOD;
}
#endif /* USE == BSD_INTERFACE */


#if USE == HPUX_INTERFACE
/* XXX untested code! */
SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  struct sctl_io hdr;
  size_t cdb_size;

  cdb_size = CDB_SIZE (*(u_char *) src);

  memset (&hdr, 0, sizeof (hdr));
  memcpy (hdr.cdb, src, src_size);
  if (dst_size && *dst_size)
    {
      assert (cdb_size == src_size);
      hdr.flags = SCTL_READ;
      hdr.data = dst;
      hdr.data_length = *dst_size;
    }
  else
    {
      assert (cdb_size <= src_size);
      hdr.data = (char *) src + cdb_size;
      hdr.data_length = src_size - cdb_size;
    }
  hdr.cdb_length = cdb_size;
  hdr.max_msecs = 60000;	/* 1 minute timeout */
  if (ioctl (fd, SIOC_IO, &hdr) < 0)
    {
      DBG(1, "sanei_scsi_cmd: ioctl(SIOC_IO) failed: %s\n",
	  strerror (errno));
      return SANE_STATUS_IO_ERROR;
    }
  if (hdr.cdb_status)
    DBG(1, "sanei_scsi_cmd: SCSI completed with cdb_status=%d\n",
	hdr.cdb_status);
  if (dst_size)
    *dst_size = hdr.data_xfer;

  if (hdr.sense_xfer > 0 && (hdr.sense[0] & 0x80) && fd_info[fd].sense_handler)
    return (*fd_info[fd].sense_handler) (fd, hdr.sense);
  return SANE_STATUS_GOOD;
}
#endif /* USE == HPUX_INTERFACE */


#if USE == OPENSTEP_INTERFACE
SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  struct scsi_req hdr;
  size_t cdb_size;

  cdb_size = CDB_SIZE (*(u_char *) src);

  memset (&hdr, 0, sizeof (hdr));
  memcpy (&hdr.sr_cdb, src, cdb_size);
  if (dst_size && *dst_size)
    {
      assert (cdb_size == src_size);
      hdr.sr_dma_dir = SR_DMA_RD;
      hdr.sr_addr = dst;
      hdr.sr_dma_max = *dst_size;
    }
  else
    {
      assert (cdb_size <= src_size);
      hdr.sr_dma_dir = SR_DMA_WR;
      hdr.sr_addr = (char *) src + cdb_size;
      hdr.sr_dma_max = src_size - cdb_size;
    }
  hdr.sr_ioto = 60;		/* I/O timeout in seconds */

  if (ioctl (fd, SGIOCREQ, &hdr) == -1)
    {
      DBG(1, "sanei_scsi_cmd: ioctl(SGIOCREQ) failed: %s\n", strerror (errno));
      return SANE_STATUS_IO_ERROR;
    }
  if (hdr.sr_io_status != 1)
    DBG(1, "sanei_scsi_cmd: SGIOCREQ completed with sr_io_status=%d\n",
	hdr.sr_io_status);
  if (hdr.sr_scsi_status == SR_IOST_CHKSV && fd_info[fd].sense_handler)
    return (*fd_info[fd].sense_handler) (fd, (u_char *) &hdr.sr_esense);
  if (dst_size)
    *dst_size = hdr.sr_dma_xfr;
  return SANE_STATUS_GOOD;
}
#endif /* USE == OPENSTEP_INTERFACE */


#if USE == DECUNIX_INTERFACE
SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  u_char sense[64];
  UAGT_CAM_CCB hdr;
  CCB_SCSIIO ccb;
  size_t cdb_size;

  cdb_size = CDB_SIZE (*(u_char *) src);

  memset (ccb, 0, sizeof (ccb));
  ccb.cam_ch.my_addr	    = (CCB_HEADER *) &ccb;
  ccb.cam_ch.cam_ccb_len    = sizeof (ccb);
  ccb.cam_ch.cam_func_code  = XPT_SCSI_IO;
  ccb.cam_ch.cam_path_id    = fd_info[fd].bus;
  ccb.cam_ch.cam_target_id  = fd_info[fd].target;
  ccb.cam_ch.cam_target_lun = fd_info[fd].lun;
  ccb.cam_ch.cam_flags	    = 0;

  if (dst_size && *dst_size)
    {
      assert (cdb_size == src_size);
      ccb.cam_ch.cam_flags |= CAM_DIR_IN;
      ccb.cam_data_ptr = (u_char *) dst;
      ccb.cam_dxfer_len = *dst_size;
    }
  else
    {
      assert (cdb_size <= src_size);
      if (cdb_size == src_size)
	ccb.cam_ch.cam_flags |= CAM_DIR_NONE;
      else
	ccb.cam_ch.cam_flags |= CAM_DIR_OUT;
      ccb.cam_data_ptr = (u_char *) src + cdb_size;
      ccb.cam_dxfer_len = src_size - cdb_size;
    }
  ccb.cam_timeout = 60;		/* set timeout in seconds */
  ccb.cam_cdb_len = cdb_size;
  memcpy (&ccb.cam_cdb_io.cam_cdb_bytes[0], src, cdb_size);

  memset (&hdr, 0, sizeof (hdr));
  hdr.uagt_ccb = (CCB_HEADER *) &ccb;
  hdr.uagt_buffer = ccb.cam_data_ptr;
  hdr.uagt_buflen = ccb.cam_dxfer_len;
  hdr.uagt_snsbuf = sense;
  hdr.uagt_snslen = sizeof (sense);
  hdr.uagt_cdb = 0;		/* indicate that CDB is in CCB */
  hdr.uagt_cdblen = 0;

  if (ioctl (cam_fd, UAGT_CAM_IO, &hdr) < 0)
    {
      DBG(1, "sanei_scsi_cmd: ioctl(UAGT_CAM_IO) failed: %s\n",
	  strerror (errno));
      return SANE_STATUS_IO_ERROR;
    }
  if (ccb.cam_ch.cam_status != CAM_REQ_CMP)
    {
      DBG(1, "sanei_scsi_cmd: UAGT_CAM_IO completed with cam_status=%d\n",
	  ccb.cam_ch.cam_status);

      if (ccb.cam_ch.cam_status == CAM_AUTOSNS_VALID
	  && fd_info[fd].sense_handler)
	return (*fd_info[fd].sense_handler) (fd, sense);
      else
	return SANE_STATUS_INVAL;
    }
  if (dst_size)
    *dst_size = ccb.cam_dxfer_len;
  return SANE_STATUS_GOOD;
}
#endif /* USE == DECUNIX_INTERFACE */


#if USE == STUBBED_INTERFACE
SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  return SANE_STATUS_UNSUPPORTED;
}
#endif /* USE == STUBBED_INTERFACE */
